package org.apache.bcel.generic;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import org.apache.bcel.classfile.Constant;
import org.apache.bcel.classfile.ConstantCP;
import org.apache.bcel.classfile.ConstantClass;
import org.apache.bcel.classfile.ConstantDouble;
import org.apache.bcel.classfile.ConstantDynamic;
import org.apache.bcel.classfile.ConstantFieldref;
import org.apache.bcel.classfile.ConstantFloat;
import org.apache.bcel.classfile.ConstantInteger;
import org.apache.bcel.classfile.ConstantInterfaceMethodref;
import org.apache.bcel.classfile.ConstantInvokeDynamic;
import org.apache.bcel.classfile.ConstantLong;
import org.apache.bcel.classfile.ConstantMethodref;
import org.apache.bcel.classfile.ConstantNameAndType;
import org.apache.bcel.classfile.ConstantPool;
import org.apache.bcel.classfile.ConstantString;
import org.apache.bcel.classfile.ConstantUtf8;
import org.apache.bcel.classfile.Utility;
import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.ObjectType;

public class ConstantPoolGen {
  private static final int DEFAULT_BUFFER_SIZE = 256;
  
  private static final String METHODREF_DELIM = ":";
  
  private static final String IMETHODREF_DELIM = "#";
  
  private static final String FIELDREF_DELIM = "&";
  
  private static final String NAT_DELIM = "%";
  
  @Deprecated
  protected int size;
  
  @Deprecated
  protected Constant[] constants;
  
  @Deprecated
  protected int index = 1;
  
  private final Map<String, Integer> stringTable = new HashMap<>();
  
  private final Map<String, Integer> classTable = new HashMap<>();
  
  private final Map<String, Integer> utf8Table = new HashMap<>();
  
  private final Map<String, Integer> natTable = new HashMap<>();
  
  private final Map<String, Integer> cpTable = new HashMap<>();
  
  public ConstantPoolGen() {
    this.size = 256;
    this.constants = new Constant[this.size];
  }
  
  public ConstantPoolGen(Constant[] cs) {
    StringBuilder sb = new StringBuilder(256);
    this.size = Math.min(Math.max(256, cs.length + 64), 65536);
    this.constants = new Constant[this.size];
    System.arraycopy(cs, 0, this.constants, 0, cs.length);
    if (cs.length > 0)
      this.index = cs.length; 
    for (int i = 1; i < this.index; i++) {
      Constant c = this.constants[i];
      if (c instanceof ConstantString) {
        ConstantString s = (ConstantString)c;
        ConstantUtf8 u8 = (ConstantUtf8)this.constants[s.getStringIndex()];
        String key = u8.getBytes();
        if (!this.stringTable.containsKey(key))
          this.stringTable.put(key, Integer.valueOf(i)); 
      } else if (c instanceof ConstantClass) {
        ConstantClass s = (ConstantClass)c;
        ConstantUtf8 u8 = (ConstantUtf8)this.constants[s.getNameIndex()];
        String key = u8.getBytes();
        if (!this.classTable.containsKey(key))
          this.classTable.put(key, Integer.valueOf(i)); 
      } else if (c instanceof ConstantNameAndType) {
        ConstantNameAndType n = (ConstantNameAndType)c;
        ConstantUtf8 u8 = (ConstantUtf8)this.constants[n.getNameIndex()];
        ConstantUtf8 u8_2 = (ConstantUtf8)this.constants[n.getSignatureIndex()];
        sb.append(u8.getBytes());
        sb.append("%");
        sb.append(u8_2.getBytes());
        String key = sb.toString();
        sb.delete(0, sb.length());
        if (!this.natTable.containsKey(key))
          this.natTable.put(key, Integer.valueOf(i)); 
      } else if (c instanceof ConstantUtf8) {
        ConstantUtf8 u = (ConstantUtf8)c;
        String key = u.getBytes();
        if (!this.utf8Table.containsKey(key))
          this.utf8Table.put(key, Integer.valueOf(i)); 
      } else if (c instanceof ConstantCP) {
        String className;
        ConstantCP m = (ConstantCP)c;
        if (c instanceof ConstantInvokeDynamic) {
          className = Integer.toString(((ConstantInvokeDynamic)m).getBootstrapMethodAttrIndex());
        } else if (c instanceof ConstantDynamic) {
          className = Integer.toString(((ConstantDynamic)m).getBootstrapMethodAttrIndex());
        } else {
          ConstantClass clazz = (ConstantClass)this.constants[m.getClassIndex()];
          ConstantUtf8 constantUtf8 = (ConstantUtf8)this.constants[clazz.getNameIndex()];
          className = Utility.pathToPackage(constantUtf8.getBytes());
        } 
        ConstantNameAndType n = (ConstantNameAndType)this.constants[m.getNameAndTypeIndex()];
        ConstantUtf8 u8 = (ConstantUtf8)this.constants[n.getNameIndex()];
        String methodName = u8.getBytes();
        u8 = (ConstantUtf8)this.constants[n.getSignatureIndex()];
        String signature = u8.getBytes();
        String delim = ":";
        if (c instanceof ConstantInterfaceMethodref) {
          delim = "#";
        } else if (c instanceof ConstantFieldref) {
          delim = "&";
        } 
        sb.append(className);
        sb.append(delim);
        sb.append(methodName);
        sb.append(delim);
        sb.append(signature);
        String key = sb.toString();
        sb.delete(0, sb.length());
        if (!this.cpTable.containsKey(key))
          this.cpTable.put(key, Integer.valueOf(i)); 
      } 
    } 
  }
  
  public ConstantPoolGen(ConstantPool cp) {
    this(cp.getConstantPool());
  }
  
  public int addArrayClass(ArrayType type) {
    return addClass_(type.getSignature());
  }
  
  public int addClass(ObjectType type) {
    return addClass(type.getClassName());
  }
  
  public int addClass(String str) {
    return addClass_(str.replace('.', '/'));
  }
  
  private int addClass_(String clazz) {
    int cpRet;
    if ((cpRet = lookupClass(clazz)) != -1)
      return cpRet; 
    adjustSize();
    ConstantClass c = new ConstantClass(addUtf8(clazz));
    int ret = this.index;
    this.constants[this.index++] = c;
    return computeIfAbsent(this.classTable, clazz, ret);
  }
  
  public int addConstant(Constant c, org.apache.bcel.generic.ConstantPoolGen cp) {
    ConstantString constantString;
    ConstantClass s;
    ConstantNameAndType n;
    ConstantCP m;
    ConstantUtf8 u8;
    ConstantClass clazz;
    ConstantUtf8 u8_2;
    ConstantNameAndType constantNameAndType1;
    ConstantUtf8 constantUtf81;
    String className, name, signature;
    Constant[] constants = cp.getConstantPool().getConstantPool();
    switch (c.getTag()) {
      case 8:
        constantString = (ConstantString)c;
        u8 = (ConstantUtf8)constants[constantString.getStringIndex()];
        return addString(u8.getBytes());
      case 7:
        s = (ConstantClass)c;
        u8 = (ConstantUtf8)constants[s.getNameIndex()];
        return addClass(u8.getBytes());
      case 12:
        n = (ConstantNameAndType)c;
        u8 = (ConstantUtf8)constants[n.getNameIndex()];
        u8_2 = (ConstantUtf8)constants[n.getSignatureIndex()];
        return addNameAndType(u8.getBytes(), u8_2.getBytes());
      case 1:
        return addUtf8(((ConstantUtf8)c).getBytes());
      case 6:
        return addDouble(((ConstantDouble)c).getBytes());
      case 4:
        return addFloat(((ConstantFloat)c).getBytes());
      case 5:
        return addLong(((ConstantLong)c).getBytes());
      case 3:
        return addInteger(((ConstantInteger)c).getBytes());
      case 9:
      case 10:
      case 11:
        m = (ConstantCP)c;
        clazz = (ConstantClass)constants[m.getClassIndex()];
        constantNameAndType1 = (ConstantNameAndType)constants[m.getNameAndTypeIndex()];
        constantUtf81 = (ConstantUtf8)constants[clazz.getNameIndex()];
        className = Utility.pathToPackage(constantUtf81.getBytes());
        constantUtf81 = (ConstantUtf8)constants[constantNameAndType1.getNameIndex()];
        name = constantUtf81.getBytes();
        constantUtf81 = (ConstantUtf8)constants[constantNameAndType1.getSignatureIndex()];
        signature = constantUtf81.getBytes();
        switch (c.getTag()) {
          case 11:
            return addInterfaceMethodref(className, name, signature);
          case 10:
            return addMethodref(className, name, signature);
          case 9:
            return addFieldref(className, name, signature);
        } 
        throw new IllegalArgumentException("Unknown constant type " + c);
    } 
    throw new IllegalArgumentException("Unknown constant type " + c);
  }
  
  public int addDouble(double n) {
    int ret;
    if ((ret = lookupDouble(n)) != -1)
      return ret; 
    adjustSize();
    ret = this.index;
    this.constants[this.index] = new ConstantDouble(n);
    this.index += 2;
    return ret;
  }
  
  public int addFieldref(String className, String fieldName, String signature) {
    int cpRet;
    if ((cpRet = lookupFieldref(className, fieldName, signature)) != -1)
      return cpRet; 
    adjustSize();
    int classIndex = addClass(className);
    int nameAndTypeIndex = addNameAndType(fieldName, signature);
    int ret = this.index;
    this.constants[this.index++] = new ConstantFieldref(classIndex, nameAndTypeIndex);
    return computeIfAbsent(this.cpTable, className + "&" + fieldName + "&" + signature, ret);
  }
  
  public int addFloat(float n) {
    int ret;
    if ((ret = lookupFloat(n)) != -1)
      return ret; 
    adjustSize();
    ret = this.index;
    this.constants[this.index++] = new ConstantFloat(n);
    return ret;
  }
  
  public int addInteger(int n) {
    int ret;
    if ((ret = lookupInteger(n)) != -1)
      return ret; 
    adjustSize();
    ret = this.index;
    this.constants[this.index++] = new ConstantInteger(n);
    return ret;
  }
  
  public int addInterfaceMethodref(MethodGen method) {
    return addInterfaceMethodref(method.getClassName(), method.getName(), method.getSignature());
  }
  
  public int addInterfaceMethodref(String className, String methodName, String signature) {
    int cpRet;
    if ((cpRet = lookupInterfaceMethodref(className, methodName, signature)) != -1)
      return cpRet; 
    adjustSize();
    int classIndex = addClass(className);
    int nameAndTypeIndex = addNameAndType(methodName, signature);
    int ret = this.index;
    this.constants[this.index++] = new ConstantInterfaceMethodref(classIndex, nameAndTypeIndex);
    return computeIfAbsent(this.cpTable, className + "#" + methodName + "#" + signature, ret);
  }
  
  public int addLong(long n) {
    int ret;
    if ((ret = lookupLong(n)) != -1)
      return ret; 
    adjustSize();
    ret = this.index;
    this.constants[this.index] = new ConstantLong(n);
    this.index += 2;
    return ret;
  }
  
  public int addMethodref(MethodGen method) {
    return addMethodref(method.getClassName(), method.getName(), method.getSignature());
  }
  
  public int addMethodref(String className, String methodName, String signature) {
    int cpRet;
    if ((cpRet = lookupMethodref(className, methodName, signature)) != -1)
      return cpRet; 
    adjustSize();
    int nameAndTypeIndex = addNameAndType(methodName, signature);
    int classIndex = addClass(className);
    int ret = this.index;
    this.constants[this.index++] = new ConstantMethodref(classIndex, nameAndTypeIndex);
    return computeIfAbsent(this.cpTable, className + ":" + methodName + ":" + signature, ret);
  }
  
  public int addNameAndType(String name, String signature) {
    int ret;
    if ((ret = lookupNameAndType(name, signature)) != -1)
      return ret; 
    adjustSize();
    int nameIndex = addUtf8(name);
    int signatureIndex = addUtf8(signature);
    ret = this.index;
    this.constants[this.index++] = new ConstantNameAndType(nameIndex, signatureIndex);
    return computeIfAbsent(this.natTable, name + "%" + signature, ret);
  }
  
  public int addString(String str) {
    int ret;
    if ((ret = lookupString(str)) != -1)
      return ret; 
    int utf8 = addUtf8(str);
    adjustSize();
    ConstantString s = new ConstantString(utf8);
    ret = this.index;
    this.constants[this.index++] = s;
    return computeIfAbsent(this.stringTable, str, ret);
  }
  
  public int addUtf8(String n) {
    int ret;
    if ((ret = lookupUtf8(n)) != -1)
      return ret; 
    adjustSize();
    ret = this.index;
    this.constants[this.index++] = new ConstantUtf8(n);
    return computeIfAbsent(this.utf8Table, n, ret);
  }
  
  protected void adjustSize() {
    if (this.index + 3 >= 65536)
      throw new IllegalStateException("The number of constants " + (this.index + 3) + " is over the size of the constant pool: " + Character.MAX_VALUE); 
    if (this.index + 3 >= this.size) {
      Constant[] cs = this.constants;
      this.size *= 2;
      this.size = Math.min(this.size, 65536);
      this.constants = new Constant[this.size];
      System.arraycopy(cs, 0, this.constants, 0, this.index);
    } 
  }
  
  private int computeIfAbsent(Map<String, Integer> map, String key, int value) {
    return ((Integer)map.computeIfAbsent(key, k -> Integer.valueOf(value))).intValue();
  }
  
  public Constant getConstant(int i) {
    return this.constants[i];
  }
  
  public ConstantPool getConstantPool() {
    return new ConstantPool(this.constants);
  }
  
  public ConstantPool getFinalConstantPool() {
    return new ConstantPool(Arrays.<Constant>copyOf(this.constants, this.index));
  }
  
  private int getIndex(Map<String, Integer> map, String key) {
    return toIndex(map.get(key));
  }
  
  public int getSize() {
    return this.index;
  }
  
  public int lookupClass(String str) {
    return getIndex(this.classTable, str.replace('.', '/'));
  }
  
  public int lookupDouble(double n) {
    long bits = Double.doubleToLongBits(n);
    for (int i = 1; i < this.index; i++) {
      if (this.constants[i] instanceof ConstantDouble) {
        ConstantDouble c = (ConstantDouble)this.constants[i];
        if (Double.doubleToLongBits(c.getBytes()) == bits)
          return i; 
      } 
    } 
    return -1;
  }
  
  public int lookupFieldref(String className, String fieldName, String signature) {
    return getIndex(this.cpTable, className + "&" + fieldName + "&" + signature);
  }
  
  public int lookupFloat(float n) {
    int bits = Float.floatToIntBits(n);
    for (int i = 1; i < this.index; i++) {
      if (this.constants[i] instanceof ConstantFloat) {
        ConstantFloat c = (ConstantFloat)this.constants[i];
        if (Float.floatToIntBits(c.getBytes()) == bits)
          return i; 
      } 
    } 
    return -1;
  }
  
  public int lookupInteger(int n) {
    for (int i = 1; i < this.index; i++) {
      if (this.constants[i] instanceof ConstantInteger) {
        ConstantInteger c = (ConstantInteger)this.constants[i];
        if (c.getBytes() == n)
          return i; 
      } 
    } 
    return -1;
  }
  
  public int lookupInterfaceMethodref(MethodGen method) {
    return lookupInterfaceMethodref(method.getClassName(), method.getName(), method.getSignature());
  }
  
  public int lookupInterfaceMethodref(String className, String methodName, String signature) {
    return getIndex(this.cpTable, className + "#" + methodName + "#" + signature);
  }
  
  public int lookupLong(long n) {
    for (int i = 1; i < this.index; i++) {
      if (this.constants[i] instanceof ConstantLong) {
        ConstantLong c = (ConstantLong)this.constants[i];
        if (c.getBytes() == n)
          return i; 
      } 
    } 
    return -1;
  }
  
  public int lookupMethodref(MethodGen method) {
    return lookupMethodref(method.getClassName(), method.getName(), method.getSignature());
  }
  
  public int lookupMethodref(String className, String methodName, String signature) {
    return getIndex(this.cpTable, className + ":" + methodName + ":" + signature);
  }
  
  public int lookupNameAndType(String name, String signature) {
    return getIndex(this.natTable, name + "%" + signature);
  }
  
  public int lookupString(String str) {
    return getIndex(this.stringTable, str);
  }
  
  public int lookupUtf8(String n) {
    return getIndex(this.utf8Table, n);
  }
  
  public void setConstant(int i, Constant c) {
    this.constants[i] = c;
  }
  
  private int toIndex(Integer index) {
    return (index != null) ? index.intValue() : -1;
  }
  
  public String toString() {
    StringBuilder buf = new StringBuilder();
    for (int i = 1; i < this.index; i++)
      buf.append(i).append(")").append(this.constants[i]).append("\n"); 
    return buf.toString();
  }
}
